import asyncio
from concurrent.futures import ThreadPoolExecutor
import contextvars
import functools
from functools import wraps
from typing import Any, Callable


def synchronous(f: Callable) -> Callable:
    """
    Run async function in place and return the result
    """

    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            asyncio.get_running_loop()

            # If running event loop already exists, run the coroutine within a executor
            executor = ThreadPoolExecutor()
            future = executor.submit(asyncio.run, f(*args, **kwargs))
            return future.result()
        except RuntimeError:
            # When no running event loop exists, create new one and run the coroutine
            try:
                return asyncio.run(f(*args, **kwargs))
            except Exception as e:
                # Suppress previous RuntimeError due to running loop
                raise e from None

    return wrapper


def asynchronous(f: Callable) -> Callable:
    """
    Replace sync function to async using aysncio thread pool
    """

    @wraps(f)
    async def wrapper(*args: Any, **kwargs: Any):
        return await _to_thread(f, *args, **kwargs)

    return wrapper


async def _to_thread(func, *args, **kwargs):
    """
    Copied from asyncio.to_thread() in Python 3.9
    """
    loop = asyncio.events.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)
    return await loop.run_in_executor(None, func_call)



from ctypes import CDLL, c_char_p, c_int, util
from enum import IntEnum
import glob
import logging
import os
from sys import platform

LOG = logging.getLogger(__name__)


DEFAULT_ENCODING: str = "utf-8"


def find_user_lib_path(libname: str):
    libpath = None
    for lib_path in os.getenv('LD_LIBRARY_PATH').split(":"):
        if platform == "linux":
            for name in glob.glob(f"{lib_path}/lib{libname}.so*"):
                libpath = name
                break
        elif platform == "darwin":
            for name in glob.glob(f"{lib_path}/lib{libname}.dylib*"):
                libpath = name
                break

    return libpath


def find_global_lib_path(libname: str):
    return util.find_library(libname)


def find_native_lib_path(libname: str):
    """Finding a native lib according to the following priority
    1. If the environment variable 'LD_LIBRARY_PATH' is set,
    this function tries to find native library found from LD_LIBRARY_PATH.
    2. Otherwise, it tries find the native library from global library paths,
    such as /usr/lib, /usr/local/lib.
    """
    libpath = None

    if os.getenv('LD_LIBRARY_PATH'):
        libpath = find_user_lib_path(libname)

    if not libpath:
        libpath = find_global_lib_path(libname)

    if not libpath:
        raise SystemExit(f'fail to find lib{libname}')

    return libpath


def __register_common_capis(ciface):
    ciface.version.argtypes = []
    ciface.version.restype = c_char_p

    ciface.git_short_hash.argtypes = []
    ciface.git_short_hash.restype = c_char_p

    ciface.build_timestamp.argtypes = []
    ciface.build_timestamp.restype = c_char_p

    ciface.register_signal_handler.argtypes = []
    ciface.register_signal_handler.restype = None

    ciface.enable_logging.argtypes = [c_int]
    ciface.enable_logging.restype = None


def find_native_libs(libname: str):
    """Finding a native lib according to the priority
    1. If the environment variable 'LD_LIBRARY_PATH' is set,
    this function tries to find native library found from LD_LIBRARY_PATH.
    2. Otherwise, it tries to find the native library embedded in the python package.
    3. If the embedded native library cannot be found,
    it tries find the native library from global library paths, such as /usr/lib, /usr/local/lib.
    """

    libpath = find_native_lib_path(libname)
    ciface = CDLL(libpath)

    if ciface:
        __register_common_capis(ciface)
        LOG.info(
            'loaded native library %s (%s %s)'
            % (
                libpath,
                ciface.version().decode(DEFAULT_ENCODING),
                ciface.git_short_hash().decode(DEFAULT_ENCODING),
            )
        )
    else:
        raise SystemExit('fail to load native library')

    return ciface


class LogLevel(IntEnum):
    """Python object correspondnig to nux_log_level_t in Nux C API"""

    OFF = 0
    ERROR = 1
    WARN = 2
    INFO = 3



from __future__ import print_function

import logging as log
import pkgutil
import sys
from typing import Optional

from packaging.version import Version


class FuriosaVersionInfo:
    def __init__(self, version: Version):
        self.version = version.base_version
        self.stage = "dev" if version.is_devrelease else "release"
        self.hash = version.local.split(".")[0][1:]

    def __str__(self):
        return f"{self.version}-{self.stage} (rev: {self.hash[0:9]})"

    def __repr__(self):
        return f"FuriosaVersionInfo(({self.stage}, {self.version}, {self.hash}))"


def get_sdk_version(module) -> Optional[FuriosaVersionInfo]:
    """Returns the git commit hash representing the current version of the application."""
    sdk_version = None
    try:
        version_string = str(pkgutil.get_data(module, 'git_version.txt'), encoding="UTF-8")
        sdk_version = FuriosaVersionInfo(Version(version_string))
    except Exception as e:  # pylint: disable=broad-except
        log.warn(e)

    return sdk_version


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)



import ctypes
import typing


class FuriosaError(Exception):
    """General exception caused by Furiosa Runtime"""

    def __init__(self, message: str):
        self._message = message

    def __repr__(self):
        return '{}'.format(self._message)

    def __str__(self):
        return self.__repr__()


def is_ok(err: typing.Union[ctypes.c_int, int]) -> bool:
    return (err.value if isinstance(err, ctypes.c_int) else err) == 0


def is_err(err: typing.Union[ctypes.c_int, int]) -> bool:
    return not is_ok(err)
